# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""
Information and methods pertaining to links and link files.

Author(s): Alec Delaney (Tekktrik)
"""

import fcntl
import json
import os
import pathlib
import shutil
from typing import Dict, List, Optional, Tuple, Union

from typer import Exit
from typing_extensions import TypeAlias

import circlink
import circlink.ledger

# Type aliases
_TableRowEntry: TypeAlias = Tuple[
    int, str, bool, pathlib.Path, pathlib.Path, bool, int, str
]


# pylint: disable=too-many-instance-attributes
class CircuitPythonLink:
    """The link to the device."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        read_path: str,
        write_path: str,
        base_dir: str,
        *,
        name: str = "",
        recursive: bool = False,
        wipe_dest: bool = False,
        skip_presave: bool = False,
        proc_id: int = 0,
        confirmed: bool = False,
        end_flag: bool = False,
        stopped: bool = False,
    ) -> None:
        """Initialize the link."""
        self._read_path = pathlib.Path(read_path)
        self._write_path = pathlib.Path(write_path)
        self._base_dir = pathlib.Path(base_dir)

        if not self._read_path.exists() and not self._read_path.parent.is_dir():
            print(
                "Read path is not valid, please reference a specific file or glob pattern for files"
            )
            raise Exit(1)

        self._name = name
        self._recursive = recursive
        self._wipe_dest = wipe_dest
        self._skip_presave = skip_presave
        self._link_id = self.get_next_link_id()
        self.process_id: int = proc_id
        self.confirmed: bool = confirmed
        self.end_flag: bool = end_flag
        self._stopped = stopped

    @staticmethod
    def get_next_link_id() -> int:
        """Get the next link ID."""
        link_gen = pathlib.Path(circlink.LINKS_DIRECTORY).glob("link*.json")
        link_nums = [int(link_file.name[4:-5]) for link_file in link_gen]
        if not link_nums:
            return 1
        return max(link_nums) + 1

    @property
    def read_path(self) -> pathlib.Path:
        """Get the read path for the link."""
        return self._read_path

    @property
    def write_path(self) -> pathlib.Path:
        """Get the write path for the link."""
        return self._write_path

    @property
    def base_dir(self) -> pathlib.Path:
        """Get the base directory for the read path."""
        return self._base_dir

    @property
    def name(self) -> str:
        """Get the link name."""
        return self._name

    @property
    def recursive(self) -> bool:
        """Whether the link is recursive for the read path."""
        return self._recursive

    @property
    def link_id(self) -> int:
        """Link ID."""
        return self._link_id

    @property
    def skip_presave(self) -> bool:
        """Whether a forced save was enacted at the start of the link."""
        return self._skip_presave

    @property
    def wipe_dest(self) -> bool:
        """Whether the wrie path was recursively wiped before starting the link."""
        return self._wipe_dest

    @property
    def stopped(self) -> bool:
        """Whether the link is marked has stopped."""
        return self._stopped

    def save_link(
        self, *, save_directory: str = circlink.LINKS_DIRECTORY
    ) -> pathlib.Path:
        """Save the link object as a file in the specified folder."""
        # Create the representative object
        link_obj = {
            "name": self._name,
            "read": str(self._read_path.resolve()),
            "write": str(self._write_path.resolve()),
            "base_dir": str(self._base_dir.resolve()),
            "recursive": self._recursive,
            "wipe_dest": self._wipe_dest,
            "skip_presave": self._skip_presave,
            "proc_id": self.process_id,
            "confirmed": self.confirmed,
            "end_flag": self.end_flag,
            "stopped": self._stopped,
        }

        # Get the filename
        save_filepath = self.link_id_to_filename(
            self._link_id, directory=save_directory
        )

        # Save the object as a JSON file
        with open(save_filepath, mode="w", encoding="utf-8") as linkfile:
            fcntl.lockf(linkfile, fcntl.LOCK_EX)
            json.dump(link_obj, linkfile, indent=4)
            fcntl.lockf(linkfile, fcntl.LOCK_UN)

        # Return the save filepath
        return pathlib.Path(save_filepath)

    @classmethod
    def load_link_by_filepath(cls, link_filepath: str) -> Optional["CircuitPythonLink"]:
        """Create a CircuitPythonLink from a JSON file, by filepath."""
        with open(link_filepath, encoding="utf-8") as linkfile:
            link_contents = linkfile.read()
            if not link_contents:
                return None
            link_obj = json.loads(link_contents)

        link = cls(
            name=link_obj["name"],
            read_path=link_obj["read"],
            write_path=link_obj["write"],
            base_dir=link_obj["base_dir"],
            recursive=link_obj["recursive"],
            wipe_dest=link_obj["wipe_dest"],
            skip_presave=link_obj["skip_presave"],
            proc_id=link_obj["proc_id"],
            confirmed=link_obj["confirmed"],
            end_flag=link_obj["end_flag"],
            stopped=link_obj["stopped"],
        )

        link._link_id = CircuitPythonLink.filename_to_link_id(link_filepath)

        return link

    @classmethod
    def load_link_by_num(cls, link_num: int) -> "CircuitPythonLink":
        """Create a CircuitPythonLink from a JSON file, by number."""
        link_filepath = cls.link_id_to_filename(link_num)
        return cls.load_link_by_filepath(link_filepath)

    @staticmethod
    def link_id_to_filename(
        num: int, *, directory: str = circlink.LINKS_DIRECTORY
    ) -> str:
        """Create a link filename from a link ID."""
        return os.path.join(directory, "link" + str(num) + ".json")

    @staticmethod
    def filename_to_link_id(filepath: Union[pathlib.Path, str]) -> int:
        """Get a link ID from a filename."""
        if isinstance(filepath, str):
            filepath = pathlib.Path(filepath)

        return int(filepath.name[4:-5])

    def get_files_monitored(self):
        """Get the existing files to be monitored by this link."""
        file_pattern = self._read_path.name
        file_parent = self._read_path.parent

        all_potential = list(
            file_parent.rglob(file_pattern)
            if self._recursive
            else file_parent.glob(file_pattern)
        )

        return [file for file in all_potential if file.is_file()]

    # pylint: disable=too-many-branches
    def begin_monitoring(self) -> None:
        """Monitor the listed file(s) for changes."""
        # Ensure the write path exists
        os.makedirs(self._write_path, exist_ok=True)

        # Wipe the destination (write path) recursively
        if self._wipe_dest:
            shutil.rmtree(self._write_path)

        # Get the files that match the read path
        read_files = self.get_files_monitored()
        update_map: Dict[pathlib.Path, float] = {}

        # Add all the files to ledger and monitor struct if not already
        for read_file in read_files:
            ledger_file_path = str(
                self.get_write_filepath(self.write_path, read_file, self.base_dir)
            )
            if circlink.ledger.append_to_ledger(
                circlink.ledger.LedgerEntry(
                    ledger_file_path, self.link_id, self.process_id
                ),
                expect_entry=False,
            ):
                update_map[read_file] = read_file.stat().st_mtime
                if not self._skip_presave:
                    self._copy_file(self._write_path, read_file, self.base_dir)

        # Initialize list for files marked for deletion
        marked_delete = []

        # Load the link and repeatedly load while not flagged to stop
        temp_link = self.load_link_by_num(self._link_id)
        while not temp_link or not temp_link.end_flag:
            # Load the link
            temp_link = self.load_link_by_num(self._link_id)

            # Detect new files
            read_files = self.get_files_monitored()
            new_files: List[pathlib.Path] = []
            for file in read_files:
                ledger_file_path = str(
                    self.get_write_filepath(self.write_path, file, self.base_dir)
                )
                if (
                    circlink.ledger.append_to_ledger(
                        circlink.ledger.LedgerEntry(
                            ledger_file_path, self.link_id, self.process_id
                        ),
                        expect_entry=False,
                    )
                    and file not in update_map
                ):
                    new_files.append(file)

            # Update the modified time for new files
            for file in new_files:
                update_map[file] = file.stat().st_mtime
                self._copy_file(self.write_path, file, self.base_dir)
            new_files = []

            # Iterate through listed existing files
            for file, last_modtime in update_map.items():
                # Detect deleted
                if not file.exists():
                    marked_delete.append(file)
                    continue

                # Detect changes
                modtime = file.stat().st_mtime
                if modtime > last_modtime:
                    self._copy_file(self._write_path, file, self.base_dir)
                    update_map[file] = modtime

            # Delete marked files
            for file in marked_delete:
                self._delete_file(self._write_path, file, self.base_dir)
                ledger_file_path = str(
                    self.get_write_filepath(self.write_path, file, self.base_dir)
                )
                ledger_entry = circlink.ledger.LedgerEntry(
                    ledger_file_path, self.link_id, self.process_id
                )
                circlink.ledger.remove_from_ledger(ledger_entry, expect_entry=True)
                try:
                    del update_map[file]
                except KeyError:
                    pass
            marked_delete = []

        # Remove files from ledger
        for file in self.get_files_monitored():
            ledger_entry = circlink.ledger.LedgerEntry(
                str(file.resolve()), self.link_id, self.process_id
            )
            circlink.ledger.remove_from_ledger(ledger_entry, expect_entry=True)

        # Mark link as end flag set and stopped, then save
        self.end_flag = True
        self._stopped = True
        self.save_link()

    @staticmethod
    def get_write_filepath(
        write_path: pathlib.Path, read_file: pathlib.Path, base_dir: pathlib.Path
    ) -> pathlib.Path:
        """Get the write filepath for a specific file."""
        read_file_relative = read_file.relative_to(base_dir)
        return write_path / read_file_relative

    @staticmethod
    def _copy_file(
        write_path: pathlib.Path,
        read_file: pathlib.Path,
        base_dir: pathlib.Path,
    ):
        read_file_relative = read_file.relative_to(base_dir)
        file_dest = write_path / read_file_relative

        if not file_dest.exists():
            os.makedirs(os.path.join(str(file_dest.parent.resolve())), exist_ok=True)

        shutil.copyfile(str(read_file.resolve()), file_dest.resolve())

    @staticmethod
    def _delete_file(
        write_path: pathlib.Path,
        read_file: pathlib.Path,
        base_dir: pathlib.Path,
    ):
        file_dest = CircuitPythonLink.get_write_filepath(
            write_path, read_file, base_dir
        )

        if file_dest.resolve().exists():
            os.remove(file_dest.resolve())


def get_links_header() -> Tuple[str, ...]:
    """Get the header row for the links list."""
    return (
        "ID",
        "Name",
        "Running?",
        "Read Path",
        "Write Path",
        "Recursive?",
        "Process ID",
        "Base Directory",
    )


def get_links_list(
    pattern: str,
    *,
    abs_paths: bool = False,
    name: str = "",
    folder: str = circlink.LINKS_DIRECTORY,
) -> List[_TableRowEntry]:
    """Get the information about links."""
    # Get the paths of all the exiting links
    link_paths = pathlib.Path(folder).glob(pattern)

    # Populate the list of links
    link_infos = []
    for link_path in link_paths:
        # Load link and start getting info
        link = CircuitPythonLink.load_link_by_filepath(str(link_path))
        link_running = not link.stopped

        # Attempt to use relative paths if possible
        if not abs_paths:
            try:
                link_read = link.read_path.resolve().relative_to(os.getcwd())
            except ValueError:
                abs_paths = True
        if abs_paths:
            link_read = link.read_path.resolve()

        # Get remaining link info
        link_write = link.write_path.resolve()

        # If not specific named link is requested, append it
        if not name or link.name == name:
            link_infos.append(
                (
                    link.link_id,
                    link.name,
                    link_running,
                    link_read,
                    link_write,
                    link.recursive,
                    link.process_id,
                    link.base_dir,
                )
            )

    # Return a sorted list of links
    return sorted(link_infos, key=lambda x: x[0])



# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

"""Configuration settings for sphinx."""

import sphinx_rtd_theme

project = "CircLink"
copyright = "2022, Alec Delaney"
author = "Alec Delaney"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ["_templates"]
exclude_patterns = [
    "sphinx.ext.autodoc",
    "sphinx.ext.todo",
]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path(), "_static"]
html_static_path = ["_static"]



# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""
The main script handling CLI interactions for ``circlink``.

Author(s): Alec Delaney (Tekktrik)
"""

import os
import shutil
import sys

import circup
import tabulate
from typer import Argument, Exit, Option, Typer

import circlink
import circlink.backend
import circlink.ledger
import circlink.link
from circlink.cli import config, workspace

# Prevent running on non-POSIX systems that don't have os.fork()
if os.name != "posix":
    print("circlink is currently only available for Linux and macOS")
    sys.exit(1)

# Create the Typer apps
app = Typer(
    add_completion=False,
    no_args_is_help=True,
    help="Autosave local files to your CircuitPython board",
)
app.add_typer(config.config_app, name="config")
app.add_typer(workspace.workspace_app, name="workspace")


# pylint: disable=too-many-arguments
@app.command()
def start(
    read_path: str = Argument(..., help="The read path/pattern of file(s) to save"),
    write_path: str = Argument(
        ...,
        help="The write path of the directory to write files, relative to the CircuitPython board",
    ),
    *,
    path: bool = Option(
        False,
        "--path",
        "-p",
        help="Designate the write path as absolute or relative to the current directory",
    ),
    name: str = Option("", "--name", "-n", help="A name for the new link"),
    recursive: bool = Option(
        False, "--recursive", "-r", help="Whether the link glob pattern is recursive"
    ),
    wipe_dest: bool = Option(
        False,
        "--wipe-dest",
        "-w",
        help="Wipe the write destination recursively before starting the link",
    ),
    skip_presave: bool = Option(
        False,
        "--skip-presave",
        "-s",
        help="Skip the inital save and write performed when opening a link",
    ),
) -> None:
    """Start a CircuitPython link."""
    circlink.backend.start_backend(
        read_path,
        write_path,
        os.getcwd(),
        path=path,
        name=name,
        recursive=recursive,
        wipe_dest=wipe_dest,
        skip_presave=skip_presave,
    )
    circlink.backend.set_cws_name("")


@app.command()
def stop(
    link_id: str = Argument(..., help="Link ID / 'last' / 'all'"),
    clear_flag: bool = Option(
        False,
        "--clear",
        "-c",
        help="Clear the history of the specified link(s) as well",
    ),
) -> bool:
    """Stop a CircuitPython link."""
    # If stopping all links, stop links using the "last" option until done
    if link_id == "all":
        link_entries = circlink.link.get_links_list("*")
        for link_entry in link_entries:
            circlink.backend.stop_backend(link_entry[0], hard_fault=False)
            if clear_flag:
                circlink.backend.clear_backend(link_entry[0], hard_fault=False)
        raise Exit()

    # If stopping the last link, calculate its link ID
    if link_id == "last":
        link_id = str(circlink.link.CircuitPythonLink.get_next_link_id() - 1)
        if link_id == "0":
            print("There are no links in the history")
            raise Exit(1)

    # Detect if the link ID is not valid
    try:
        link_id = int(link_id)
    except ValueError as err:
        print('Link ID must be the ID, "last", or "all"')
        raise Exit(1) from err

    # Stop the link, clear as well if requested
    circlink.backend.stop_backend(link_id)
    if clear_flag:
        circlink.backend.clear_backend(link_id)


@app.command()
def clear(
    link_id: str = Argument(..., help="Link ID / 'last' / 'all'"),
    *,
    force: bool = Option(
        False, "--force", "-f", help="Ignore warning and force clear from history"
    ),
) -> None:
    """Clear the link from the history."""
    # If clearing all links, repetitively clear the last link
    if link_id == "all":
        link_entries = circlink.link.get_links_list("*")
        for link_entry in link_entries:
            circlink.backend.clear_backend(link_entry[0], force=force, hard_fault=False)
        raise Exit()

    # If clearing the last link link, calculate its link ID
    if link_id == "last":
        link_id = str(circlink.link.CircuitPythonLink.get_next_link_id() - 1)
        if link_id == "0":
            return

    # Detect if the link ID is not valid
    try:
        link_id = int(link_id)
    except ValueError as err:
        print('Link ID must be the ID, "last", or "all"')
        raise Exit(1) from err

    # Clear the link
    circlink.backend.clear_backend(link_id, force=force)


@app.command()
def view(
    link_id: str = Argument("all", help="Link ID / 'last' / 'all' (default)"),
    *,
    abs_paths: bool = Option(
        False, "--abs-path", "-a", help="Show the read path as absolute"
    ),
) -> None:
    """List links in the history."""
    # For recursion purposes, note whether link ID is "last"
    last_requested_flag = False

    # Handle cases of link ID being "all" or "last", or just parse link ID
    if link_id == "all":
        pattern = "*"
    elif link_id == "last":
        last_requested_flag = True
        link_id = str(circlink.link.CircuitPythonLink.get_next_link_id() - 1)
        pattern = "link" + link_id + ".json"
        if link_id == "0":
            pattern = "*"
    else:
        try:
            int(link_id)
            pattern = "link" + link_id + ".json"
        except ValueError:
            print('Please use a valid link ID, "last", or "all" (default)')

    # Discard the link base directory for printing purposes
    link_infos = circlink.backend.view_backend(pattern, abs_paths=abs_paths)

    # Handle if no links available
    if not link_infos:
        if link_id == "all" or last_requested_flag:
            print("No links in the history to view")
            raise Exit()
        print("This link ID is not in the history")
        raise Exit(1)


@app.command()
def restart(link_id: str = Argument(..., help="Link ID / 'last' / 'all'")) -> None:
    """Restart a link."""
    # Handle cases of "all" or "last", or parse link ID
    if link_id == "all":
        pattern = "*"
    elif link_id == "last":
        link_id = str(circlink.link.CircuitPythonLink.get_next_link_id() - 1)
        pattern = "link" + link_id + ".json"
        if link_id == "0":
            pattern = "*"
    else:
        try:
            int(link_id)
            pattern = "link" + link_id + ".json"
        except ValueError as err:
            print('Please use a valid link ID, "last", or "all"')
            raise Exit(1) from err

    # Get the list of links in history if possible
    link_list = circlink.link.get_links_list(pattern)
    if not link_list:
        print("There are no links in the history to restart")
        raise Exit(1)

    # Attempt to restart and clear the link if it's not active
    for link in link_list:
        if link[2]:
            print(f"Link #{link[0]} is active, not restarting this link.")
        else:
            circlink.backend.start_backend(
                str(link[3]),
                str(link[4]),
                link[-1],
                name=link[1],
                recursive=link[5],
                path=True,
            )
            circlink.backend.clear_backend(link[0])


@app.command()
def detect() -> None:
    """Attempt to detect a CircuitPython board."""
    device = circup.find_device()
    if device:
        print("CircuitPython device detected:", device)
    else:
        print("No CircuitPython device detected")


def about_cb() -> None:
    """Display information about circlink."""
    print("Originally built with love by Tekktrik")
    print("Happy hackin'!")
    raise Exit()


def version_cb() -> None:
    """Display the current version of circlink."""
    print(circlink.__version__)
    raise Exit()


@app.callback(invoke_without_command=True)
def callback(
    version: bool = Option(
        False, "--version", "-v", help="Display the version of circlink"
    ),
    about: bool = Option(False, "--about", "-a", help="A bit about circlink"),
    reset: bool = Option(
        False, "--reset", help="Reset the circlink configuration settings"
    ),
) -> None:
    """Display the current version of circlink."""
    circlink.ensure_app_folder_setup()

    if version:
        version_cb()
    if about:
        about_cb()
    if reset:
        reset_cb()


def reset_cb() -> None:
    """
    Reset the app directory.

    Useful if you upgrade circlink and there are breaking changes.
    """
    shutil.rmtree(circlink.APP_DIRECTORY)
    print("Removed circlink app directory, settngs and history deleted!")
    print("These will be created on next use of circlink.")
    print("Please check the integrity of any files handled by circlink.")
    raise Exit()


@app.command()
def ledger() -> None:
    """View the ledger of files controlled by links."""
    # Get the list of ledger entries if possible
    ledger_entries = list(circlink.ledger.iter_ledger_entries())
    if not ledger_entries:
        print("No files being tracked by circlink")
        raise Exit()

    # Display the process ID of links depending on settings
    table_headers = ("Write Path", "Link")
    if circlink.get_settings()["display"]["info"]["process-id"]:
        table_headers = table_headers + ("Process ID",)
    else:
        ledger_entries = [entry[:-1] for entry in ledger_entries]

    # Print the table with the format specified in config settings
    print(
        tabulate.tabulate(
            ledger_entries,
            headers=table_headers,
            tablefmt=circlink.get_settings()["display"]["table"]["format"],
        )
    )



# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""
Backend operations for circlink.

Author(s): Alec Delaney (Tekktrik)
"""

import datetime
import os
import signal
from typing import Iterable, List

import circup
import psutil
import tabulate
from typer import Exit

import circlink
import circlink.ledger
import circlink.link


# pylint: disable=too-many-locals,too-many-branches,too-many-arguments
def start_backend(
    read_path: str,
    write_path: str,
    base_dir: str,
    *,
    path: bool = False,
    name: str = "",
    recursive: bool = False,
    wipe_dest: bool = False,
    skip_presave: bool = False,
) -> None:
    """Start a link (backend)."""
    # Only allow recursive setting with glob patterns
    if "*" not in read_path and recursive:
        print("--recursive can only be used with glob patterns!")
        raise Exit(1)

    # Attempt to find the CircuitPython board unless explicitly set otherwise
    if not path:
        device_path = circup.find_device()
        if not device_path:
            print("Cound not auto-detect board path!")
            raise Exit(1)
        write_path = os.path.join(device_path, write_path)

    # Warn if the board (or write path in general) is not write accessible
    if not os.access(write_path, os.W_OK):
        print("Cannot write to the device or specified path")
        print("If using CircuitPython board, please ensure it is mounted")
        raise Exit(1)

    # Set the base directory is current directory if not provided
    base_dir = os.getcwd() if not base_dir else base_dir

    # Create the link object
    link = circlink.link.CircuitPythonLink(
        read_path,
        write_path,
        base_dir,
        name=name,
        recursive=recursive,
        wipe_dest=wipe_dest,
        skip_presave=skip_presave,
    )

    # Get the link ID and save the link
    link_id = link.link_id
    link.save_link()

    # Detect whether bad read path/pattern provided
    try:
        link.read_path.resolve().relative_to(base_dir)
    except ValueError as err:
        print(
            "Error occurred, please ensure the read file/pattern "
            "is relative to the current directory"
        )
        raise Exit(1) from err

    # Fork the process to begin start the link
    pid = os.fork()  # pylint: disable=no-member

    if pid:  # Current process, pid is that of spawned process
        # Save the link with the process ID
        link.process_id = pid
        link.save_link()

        # Attempt to wait for spawned process to be confirmed
        start_time = datetime.datetime.now()
        error_time = start_time + datetime.timedelta(seconds=5)
        while not link or not link.confirmed:
            link = circlink.link.CircuitPythonLink.load_link_by_num(link_id)
            if datetime.datetime.now() >= error_time:
                try:
                    os.kill(pid, signal.SIGTERM)
                except ProcessLookupError:
                    pass
                raise OSError("Could not start link process!")

        print(f"Started link #{link.link_id}")

    else:  # Spawned process, PID is 0
        # Wait for the process ID to be avaiable
        while not link or not link.process_id:
            link = circlink.link.CircuitPythonLink.load_link_by_num(link_id)

        # Mark the link is confirmed and save it
        link.confirmed = True
        link.save_link()

        # Begin monitoring files, exiting successfully when completed or error if needed
        try:
            link.begin_monitoring()
        except FileNotFoundError:
            for file in link.get_files_monitored():
                ledger_entry = circlink.ledger.LedgerEntry(
                    str(file.resolve()), link.link_id, link.process_id
                )
                circlink.ledger.remove_from_ledger(ledger_entry)
            Exit(1)
        raise Exit()


def stop_backend(link_id: int, *, hard_fault: bool = True) -> bool:
    """Stop a link (backend)."""
    # Attempt to get the link by ID
    try:
        link = circlink.link.CircuitPythonLink.load_link_by_num(link_id)
    except FileNotFoundError as err:
        print(f"Link #{link_id} does not exist!")
        raise Exit(1) from err

    # Announce if the link is already stopped
    if link.stopped:
        print(f"Link #{link.link_id} is already stopped")
        if hard_fault:
            raise Exit()
        return False

    # Detect whether the link exists and is circlink
    try:
        circlink_process = psutil.Process(link.process_id)
        maybe_process = circlink_process.name() == "circlink"
    except psutil.NoSuchProcess:
        maybe_process = False

    if not maybe_process:
        print(
            f"Problem encountered stopping link #{link_id}!\n"
            "Asscoiated proess either does not exist, was already "
            "stopped, or isn't circlink.\n"
            "Consider using the clear command with the --force flag to "
            "clear it from the history."
        )
        raise Exit(1)

    # Mark the link for termination
    link.end_flag = True
    link.save_link()

    # Wait for confirmation that the link has stopped
    start_time = datetime.datetime.now()
    error_time = start_time + datetime.timedelta(seconds=5)
    while not link or not link.stopped:
        link = circlink.link.CircuitPythonLink.load_link_by_num(link_id)
        if datetime.datetime.now() >= error_time:
            print(f"Link #{link.link_id} could not be stopped!")
            if hard_fault:
                raise Exit(1)
            return False

    # Announce the link has stopped
    print(f"Stopped link #{link_id}")
    return True


def clear_backend(
    link_id: int, *, force: bool = False, hard_fault: bool = False
) -> bool:
    """Clear a link (backend)."""
    # Get the link object by link ID
    try:
        link = circlink.link.CircuitPythonLink.load_link_by_num(link_id)
    except FileNotFoundError as err:
        print(f"Link #{link_id} does not exist!")
        raise Exit(1) from err

    # If the link is not marked as stop, announce "--force" is needed
    if not link.stopped and not force:
        print("Can only clear links marked as inactive.")
        print(f"To force clear link #{link.link_id}, use the --force option.")
        if hard_fault:
            raise Exit(1)
        return False

    # Remove the file associated with the link
    os.remove(link.link_id_to_filename(link_id))
    print(f"Removed link #{link_id} from history")

    # Remove file from ledger, just in case
    for entry in circlink.ledger.iter_ledger_entries():
        if entry.link_id == link_id:
            circlink.ledger.remove_from_ledger(entry, expect_entry=True, use_lock=False)

    set_cws_name("")

    return True


def retrieve_links_info(
    pattern: str = "*",
    *,
    abs_paths: bool = False,
    folder: str = circlink.LINKS_DIRECTORY,
    exclude: Iterable[str] = ("Base Directory",),
) -> List[tuple]:
    """Retrieve information about a collection of links."""
    # Discard the link base directory for printing purposes
    show_list = list(circlink.link.get_links_header())
    link_infos = list(
        circlink.link.get_links_list(pattern, abs_paths=abs_paths, folder=folder)
    )

    # Remove unwanted columns
    new_link_infos = []
    for entry in link_infos:
        new_entry = ()
        for exculde_index, exclude_col in enumerate(show_list):
            if exclude_col not in exclude:
                new_entry: tuple = new_entry + (entry[exculde_index],)
        new_link_infos.append(new_entry)

    return new_link_infos


def view_backend(
    pattern: str = "*",
    *,
    abs_paths: bool = False,
    folder: str = circlink.LINKS_DIRECTORY,
    exclude: Iterable[str] = ("Base Directory",),
) -> None:
    """View a collection of links (backend)."""
    show_list = list(circlink.link.get_links_header())
    for exclude_header in exclude:
        show_list.remove(exclude_header)
    show_list = tuple(show_list)
    if (
        not circlink.get_settings()["display"]["info"]["process-id"]
        and "Process ID" not in exclude
    ):
        exclude = exclude + ("Process ID",)
        show_list = list(show_list)
        show_list.remove("Process ID")
        show_list = tuple(show_list)

    # Discard the link base directory for printing purposes
    link_infos = retrieve_links_info(
        pattern, abs_paths=abs_paths, folder=folder, exclude=exclude
    )

    # Print the table with the format based on config settings
    if link_infos:
        print(
            tabulate.tabulate(
                link_infos,
                headers=show_list,
                tablefmt=circlink.get_settings()["display"]["table"]["format"],
            )
        )

    return link_infos


def get_cws_name() -> str:
    """Get the current workspace name."""
    with open(circlink.CURRENT_WORKSPACE_FILE, encoding="utf-8") as cwsfile:
        name = cwsfile.read()
    return None if not name else name


def set_cws_name(name: str) -> None:
    """Set the current workspace name."""
    with open(circlink.CURRENT_WORKSPACE_FILE, mode="w", encoding="utf-8") as cwsfile:
        cwsfile.write(name)



# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""
Information and methods pertaining to the ledger file.

Author(s): Alec Delaney (Tekktrik)
"""

import csv
import fcntl
import functools
from collections import namedtuple
from typing import Iterator, Literal, Optional

import circlink

# Namedtuple for ledger entries
LedgerEntry = namedtuple("LedgerEntry", ("filename", "link_id", "process_id"))


def with_ledger(mode: str = "a"):
    """
    Use the ledger file.

    Manages locking and unlocking the file
    """

    def decorator_with_ledger(func):
        """Work with the ledger file."""

        @functools.wraps(func)
        def wrapper_with_ledger(
            entry: LedgerEntry,
            *,
            expect_entry: Optional[bool] = None,
            use_lock: bool = True,
        ) -> bool:
            """Edit the ledger."""
            # Open the ledger file
            with open(circlink.LEDGER_FILE, mode=mode, encoding="utf-8") as filedesc:
                # Use a file lock if requested
                if use_lock:
                    fcntl.lockf(filedesc, fcntl.LOCK_EX)

                # Handle the entry
                if (expect_entry is None) or (
                    expect_entry == (entry.filename in iter_ledger_filenames())
                ):
                    result = func(entry, filedesc=filedesc)
                else:
                    result = False

                # Release the file lock if needed
                if use_lock:
                    fcntl.lockf(filedesc, fcntl.LOCK_UN)

                return result

        return wrapper_with_ledger

    return decorator_with_ledger


@with_ledger(mode="a")
def append_to_ledger(entry: LedgerEntry, **args) -> Literal[True]:
    """
    Add a file to the ledger.

    Returns whether the file actually was added (True) or if it already
    existed (False).
    """
    csvwriter = csv.writer(args["filedesc"])
    csvwriter.writerow(entry)
    return True


@with_ledger(mode="w")
def remove_from_ledger(entry: LedgerEntry, **args) -> Literal[True]:
    """
    Remove a file from the ledger.

    Returns whether the file actually was removed (True) or if it didn't
    exist (False).
    """
    csvwriter = csv.writer(args["filedesc"])
    for existing_entry in iter_ledger_filenames(False):
        if existing_entry != entry:
            csvwriter.writerow(existing_entry)
    return True


def iter_ledger_entries(use_lock: bool = True) -> Iterator[LedgerEntry]:
    """Iterate through ledger entries."""
    with open(circlink.LEDGER_FILE, mode="r+", encoding="utf-8") as csvfile:
        if use_lock:
            fcntl.lockf(csvfile, fcntl.LOCK_EX)
        csvreader = csv.reader(csvfile)
        for ledger_entry in csvreader:
            yield LedgerEntry(
                ledger_entry[0], int(ledger_entry[1]), int(ledger_entry[2])
            )
        if use_lock:
            fcntl.lockf(csvfile, fcntl.LOCK_UN)


def iter_ledger_filenames(use_lock: bool = True) -> Iterator[str]:
    """Iterate through ledger entry filenames."""
    for entry in iter_ledger_entries(use_lock):
        yield entry.filename



# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""
The sub-script for handling the config options for ``circlink``.

Author(s): Alec Delaney (Tekktrik)
"""

import json
import os

import yaml
from typer import Argument, Exit, Option, Typer

import circlink

config_app = Typer(
    add_completion=False,
    no_args_is_help=True,
    help="Change config settings for circlink",
)


@config_app.callback(invoke_without_command=True)
def callback(
    filepath: bool = Option(
        False, "--filepath", "-f", help="Print the settings file location"
    ),
    reset: bool = Option(
        False, "--reset", help="Reset the configuration settings to their defaults"
    ),
) -> None:
    """Run the callback for the config subcommand."""
    if filepath:
        print(f"Settings file: {os.path.abspath(circlink.SETTINGS_FILE)}")
        raise Exit()
    if reset:
        circlink.reset_config_file()


@config_app.command()
def view(
    config_path: str = Argument("all", help="The setting to view, using dot notation")
) -> None:
    """View a config setting for circlink."""
    # Get the settings, show all settings if no specific on is specified
    setting = circlink.get_settings()
    if config_path == "all":
        print(json.dumps(setting, indent=4))
        raise Exit()

    # Get the specified settings
    config_args = config_path.split(".")
    try:
        for extra_arg in config_args[:-1]:
            setting = setting[extra_arg]
        value = setting[config_args[-1]]
    except KeyError as err:
        print(f"Setting {config_path} does not exist")
        raise Exit(1) from err

    # Show the specified setting
    print(f"{config_path}: {json.dumps(value, indent=4)}")


@config_app.command()
def edit(
    config_path: str = Argument("all", help="The setting to view, using dot notation"),
    value: str = Argument(..., help="The value to set for the setting"),
) -> None:
    """Edit a config setting for circlink."""
    # Get the settings, use another reference to parse
    orig_setting = circlink.get_settings()
    setting = orig_setting
    config_args = config_path.split(".")

    # Handle bool conversions
    if value.lower() == "true":
        value = True
    elif value.lower() == "false":
        value = False

    # Attempt to parse for the specified config setting and set it
    try:
        for extra_arg in config_args[:-1]:
            setting = setting[extra_arg]
        prev_value = setting[config_args[-1]]
        prev_value_type = type(prev_value)
        if prev_value_type == dict:
            raise ValueError
        if prev_value_type == bool and value not in (True, False):
            raise TypeError
        setting[config_args[-1]] = prev_value_type(value)
    except KeyError as err:
        print(f"Setting {config_path} does not exist")
        raise Exit(1) from err
    except TypeError as err:
        print(
            f"Cannot use that value for this setting, must be of type {prev_value_type}"
        )
        raise Exit(1) from err
    except ValueError as err:
        print("Cannot change this setting, please change the sub-settings within it")
        raise Exit(1) from err

    # Write the settings back to the file
    with open(circlink.SETTINGS_FILE, mode="w", encoding="utf-8") as yamlfile:
        yaml.safe_dump(orig_setting, yamlfile)



# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""
The sub-script for handling the workspace options for ``circlink``.

Author(s): Alec Delaney (Tekktrik)
"""

import os
import pathlib
import shutil
import zipfile
from typing import Dict, Optional

from typer import Argument, Exit, Option, Typer

import circlink
import circlink.backend
import circlink.link

workspace_app = Typer(
    add_completion=False,
    no_args_is_help=True,
    help="Save and load workspace settings.",
)


def get_workspaces() -> Dict[str, pathlib.Path]:
    """Iterate through the workspace names and folder filepaths."""
    workspaces = [
        path
        for path in pathlib.Path(circlink.WORKSPACE_LIST_DIRECTORY).glob("*")
        if path.is_dir()
    ]
    return {path.name: path for path in workspaces}


def _remove_workspace(name: str) -> None:
    """Remove a saved workspace (if it exists)."""
    for ws_name, ws_path in get_workspaces().items():
        if name == ws_name:
            shutil.rmtree(str(ws_path.resolve()))

    circlink.backend.set_cws_name("")


def _ensure_new_workspace(name: str, raise_error: bool = True) -> bool:
    """Save a workspace (backend)."""
    if name in get_workspaces():
        if raise_error:
            print(f"Cannot use name {name} - workspace already saved with this name.")
            raise Exit(1)
        return False
    return True


def _get_ws_path(name: str) -> str:
    """Get the for a workspace directory name."""
    return os.path.join(circlink.WORKSPACE_LIST_DIRECTORY, name)


@workspace_app.command()
def current() -> None:
    """Get the current workspace name."""
    name = circlink.backend.get_cws_name()
    if not name:
        print("Current workspace is not named")
        raise Exit()
    print(name)


@workspace_app.command(name="list")
def workspace_list() -> None:
    """List all existing workspaces."""
    workspaces = get_workspaces()
    if workspaces:
        for workspace in get_workspaces():
            print_text = (
                "* " + workspace
                if workspace == circlink.backend.get_cws_name()
                else workspace
            )
            print(print_text)
    else:
        print("No workspaces saved")


@workspace_app.command()
def delete(name: str = Argument(..., help="Name of the workspace to delete")) -> None:
    """Delete a workspace."""
    if name not in get_workspaces():
        print(f"Workspace '{name}' does not exist")
        raise Exit(1)

    _remove_workspace(name)
    print(f"Workspace '{name}' deleted")


@workspace_app.command()
def rename(
    old_name: str = Argument(..., help="Name of the workspace to rename"),
    new_name: str = Argument(..., help="New name for the workspace"),
) -> None:
    """Rename a workspace."""
    _ensure_new_workspace(new_name)

    old_path = _get_ws_path(old_name)

    parent = os.path.dirname(old_path)
    new_path = os.path.join(parent, new_name)

    os.rename(old_path, new_path)

    if circlink.backend.get_cws_name() == old_name:
        circlink.backend.set_cws_name(new_name)

    print(f"Workspace '{old_name}' renamed to '{new_name}'")


@workspace_app.command()
def save(
    name: str = Argument(..., help="The name of the new workspace"),
    *,
    overwrite: bool = Option(
        False,
        "--overwrite",
        "-o",
        help="Whether to overwrite an existing workspace with the same name",
    ),
) -> None:
    """Save the current link state as a workspace."""
    if not circlink.link.get_links_list("*"):
        print("No links are in the history, nothing to save")
        raise Exit(1)

    if not overwrite:
        _ensure_new_workspace(name)

    _remove_workspace(name)

    new_ws_folder = _get_ws_path(name)
    os.mkdir(new_ws_folder)
    links_path = pathlib.Path(circlink.LINKS_DIRECTORY)

    for link_index, link_path in enumerate(links_path.glob("*")):
        link = circlink.link.CircuitPythonLink.load_link_by_filepath(str(link_path))
        link._link_id = link_index + 1  # pylint: disable=protected-access
        link.process_id = 0
        link.end_flag = True
        link.confirmed = True
        link._stopped = True  # pylint: disable=protected-access
        link.save_link(save_directory=new_ws_folder)

    circlink.backend.set_cws_name(name)

    print(f"New workspace saved as '{name}'")


@workspace_app.command()
def load(name: str = Argument(..., help="Name of the workspace to load")) -> None:
    """Load a workspace."""
    links_folder = os.path.join(circlink.LINKS_DIRECTORY)
    links_path = pathlib.Path(links_folder)
    ws_path = pathlib.Path(circlink.WORKSPACE_LIST_DIRECTORY) / name

    if list(links_path.glob("*")):
        print("Cannot load workspace with files in the history.")
        print("Please clear the history with the clear command.")
        raise Exit(1)

    if _ensure_new_workspace(name, raise_error=False):
        print("This workspace does not exist!")
        raise Exit(1)

    for link_path in ws_path.glob("*"):
        link = circlink.link.CircuitPythonLink.load_link_by_filepath(str(link_path))
        link.save_link()

    circlink.backend.set_cws_name(name)

    print(f"Loaded workspace '{name}'")


@workspace_app.command()
def view(name: str = Argument(..., help="The name of the workspace to view")):
    """View details about a workspace."""
    results = circlink.backend.view_backend(
        "*",
        folder=_get_ws_path(name),
        exclude=("Base Directory", "Running?", "Process ID"),
    )
    if not results:
        print(f"Workspace '{name}' does not exist!")
        raise Exit(1)


@workspace_app.command()
def export(
    name: str = Argument(..., help="Name of the workspace to export"),
    path: str = Argument(
        ..., help="The folder where the workspace will be exported to"
    ),
) -> None:
    """Export a workspace."""
    export_folder = _get_ws_path(name)
    export_paths = [path.resolve() for path in pathlib.Path(export_folder).glob("*")]
    exported_path = os.path.join(path, name + ".zip")

    with zipfile.ZipFile(
        exported_path, mode="w", compression=zipfile.ZIP_DEFLATED
    ) as zip_file:
        for export_path in export_paths:
            zip_file.write(str(export_path), arcname=export_path.name)


@workspace_app.command(name="import")
def workspace_import(
    filepath: str = Argument(..., help="Filepath of the packaged workspace"),
    name: Optional[str] = Option(
        None, "--name", "-n", help="A name to give to the imported workspace"
    ),
) -> None:
    """Import a workspace."""
    package_name = os.path.basename(filepath)[:-4] if not name else name
    package_ext = os.path.splitext(filepath)[1]

    if package_ext.lower() != ".zip":
        print("Imported workspaces must be ZIP files")
        raise Exit(1)

    _ensure_new_workspace(package_name)
    new_ws_path = _get_ws_path(package_name)
    os.mkdir(new_ws_path)

    with zipfile.ZipFile(filepath) as zip_file:
        zip_file.extractall(new_ws_path)



# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""
The main script handling CLI interactions for ``circlink``.

Author(s): Alec Delaney (Tekktrik)
"""

import os
import pathlib
import shutil

import typer
import yaml

__version__ = "1.2.3"

# Filepath constants
APP_DIRECTORY = typer.get_app_dir("circlink")
LINKS_DIRECTORY = os.path.join(APP_DIRECTORY, "links")
LEDGER_FILE = os.path.join(APP_DIRECTORY, "ledger.csv")
SETTINGS_FILE = os.path.join(APP_DIRECTORY, "settings.yaml")
WORKSPACE_DIRECTORY = os.path.join(APP_DIRECTORY, "workspaces")
WORKSPACE_LIST_DIRECTORY = os.path.join(WORKSPACE_DIRECTORY, "saved")
CURRENT_WORKSPACE_FILE = os.path.join(WORKSPACE_DIRECTORY, "current.txt")


def get_settings():
    """Get the contents of the settings file."""
    with open(SETTINGS_FILE, encoding="utf-8") as yamlfile:
        return yaml.safe_load(yamlfile)


def reset_config_file() -> None:
    """Reset the config file."""
    settings_file = os.path.join(__file__, "..", "templates", "settings.yaml")
    shutil.copy(os.path.abspath(settings_file), SETTINGS_FILE)


def ensure_app_folder_setup() -> None:
    """Ensure that the configuration folder exists."""
    if not os.path.exists(APP_DIRECTORY):
        os.mkdir(APP_DIRECTORY)

    ensure_links_folder()
    ensure_ledger_file()
    ensure_settings_file()
    ensure_workspace_tree()


def ensure_settings_file() -> None:
    """Ensure the settings file is set up."""
    settings_path = pathlib.Path(SETTINGS_FILE)
    if not settings_path.exists():
        reset_config_file()


def ensure_links_folder() -> None:
    """Ensure the links folder is created."""
    if not os.path.exists(LINKS_DIRECTORY):
        os.mkdir(LINKS_DIRECTORY)


def ensure_ledger_file() -> None:
    """Ensure the ledger file exists, or create it if not."""
    ledger_path = pathlib.Path(LEDGER_FILE)
    ledger_path.touch(exist_ok=True)


def ensure_workspace_tree() -> None:
    """Ensure the workspace tree exists, or create it if not."""
    for directory in (WORKSPACE_DIRECTORY, WORKSPACE_LIST_DIRECTORY):
        if not os.path.exists(directory):
            os.mkdir(directory)
    pathlib.Path(CURRENT_WORKSPACE_FILE).touch(exist_ok=True)
